/*
 * Created on Dec 17, 2004
 *
 */
package freq;

import javax.swing.JOptionPane;

import math.Log10;
import model.Model;
import model.Parameter;

import org.jscience.mathematics.matrices.LUDecomposition;
import org.jscience.mathematics.matrices.Matrix;
import org.jscience.mathematics.numbers.Complex;
import org.lsmp.djep.djep.DJep;
import org.nfunk.jep.Node;

import sbml.SBML_Reader;
import sbw.Jarnac;

/**
 * @author Klaus Maier
 * For any comments or bugs please refer to klaus.maier@gmx.li
 *	
 *	This class is used to compute the frequency response of a given model. 
 *	
 *	Original description of the algorithms can be found in
 *	Ingalls BP: A Frequency Domain Approach to Sensitivity Analysis of Biochemical Networks; Journal of Physical Chemistry B108, 1143-1152
 *	
 *	
 */
public class FrequencyResponse 
{
	private final Complex i = Complex.I;  // the imaginary unit
	
	Model model;   // the model for which the frequency response shall be computed
		
	Matrix jac_identity;  // the identity matrix with as many rows as in the Jacobian
	Matrix rate_identity;  // the identity matrix with as many rows as there are fluxes 
	Matrix jacobian;   // the jacobian at steady state
	Matrix deltaV_deltaP;  // the reaction rates differentiated by the parameters of interest
	Matrix Nr;	// the reduced stoichiometric matrix
	Matrix Nr_times_deltaV_deltaP;  // the reduced stoichiometric matrix Nr multiplied with the matrix "deltaV_deltaP"
	Matrix deltaV_deltaS_times_L;   // the Product from the Linker Matrix L multiplied with the derivatives of the 
	                                // reaction rates with respect to the species
	Matrix DsiInverse;  // diagonal matrix with the steady state concentrations of the independent species (inverse)
	Matrix DvInverse;  // diagonal matrix with the steady state rates (inverse)
	Matrix Dv;  // diagonal matrix with the steady state rates
	Matrix Dp;  // diagonal matrix with the steady state paramter values
		
	double[][] spec_magnitude;
	double[][] spec_phase;
	double[][] rate_magnitude;
	double[][] rate_phase;
	
	double[] frequency;
	
	Jarnac jarnac;
		
	public void setModel(String sbml)
	{
		jarnac = new Jarnac();
		
		try 
		{
			model = SBML_Reader.load(sbml);
			jarnac.loadSBML(sbml);
			initJacobian();		
		} 
		catch (Exception e) 
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * calculates the Jacobian for the steady state values.
	 */
	private void initJacobian()
	{
		jarnac.calcSteadyState();
		// now get the Jacobian at steady state
		double[][] jac = jarnac.getJacobian();
		// store the values in the Matrix jacobian
		Complex[][] jacVals = new Complex[jac.length][jac[0].length];
		for (int i = 0; i < jacVals.length; i++) 		
			for (int ii = 0; ii < jacVals[i].length; ii++) 			
				jacVals[i][ii] = Complex.valueOf(jac[i][ii],0);		
		jacobian = Matrix.valueOf(jacVals);  // get new matrix with these values
	}
	/**
	 * used to differentiate the reaction rates with respect 
	 * to the user-defined parameters
	 * 
	 * @param parameters
	 * @throws Exception
	 * @throws NumberFormatException
	 */
	private void init_DeltaV_DeltaP(String parameter) throws NumberFormatException, Exception
	{
		DJep djep;
		Node node;
		double dval;
		double[] ssVals;
		String[] species;		
		String[] reactionRates = model.get_reactionRates();  // get the reaction rates
		Complex[][] vals = new Complex[reactionRates.length][1];  // initialize an empty array
		
		// initialise the jep Parser
		djep = new DJep();
		djep.addStandardConstants();
		djep.addStandardFunctions();
		djep.addComplex();
		djep.setAllowUndeclared(true);
		djep.setAllowAssignment(true);
		
		// now assign to all possible unknowns the proper double value
		for (int i = 0; i < model.getAllParameter().size(); i++)  // start with the parameters
			djep.addVariable(((Parameter)model.getAllParameter().get(i)).getName(), 
					((Parameter)model.getAllParameter().get(i)).getValue());
		
		species = jarnac.getSpeciesNames();
		ssVals = jarnac.getSteadyState();
		
		for (int i = 0; i < species.length; i++)
			djep.addVariable(species[i], 
					ssVals[i]);
		
		// now calculate the derivatives symbolically
		for (int i = 0; i < vals.length; i++)
		{	
			for (int ii = 0; ii < vals[i].length; ii++) 
			{
				node = djep.parse(model.get_reactionRates()[i]);		// pass the reactionRate to the djep-parameter				
				node = djep.differentiate(node,parameter);   // differentiate
								
				dval = Double.parseDouble(djep.evaluate(node)+"");    // evaluate and store result in "val"
				vals[i][ii] = Complex.valueOf(dval,0);   // save as complex number z = "val" + 0*i				
			}
		}				
		// Now we can initiate the desired matrix
		deltaV_deltaP = Matrix.valueOf(vals);
	}

	/**
	 * Initiates the matrix Nr, i.e. the reduced stoichiometric matrix
	 *
	 */
	private void init_Nr()
	{
		// store Nr in the matrix "Nr"
		// use complex values
		Complex[][] complex = new Complex[model.getNr().getRowDimension()][model.getNr().getColumnDimension()];
		for (int i = 0; i < complex.length; i++) 
			for (int ii = 0; ii < complex[i].length; ii++) 
				complex[i][ii] = Complex.valueOf(model.getNr().get(i,ii), 0);
		Nr = Matrix.valueOf(complex);
	}
	
	/** 
	 * Initiates the matrix "Nr_times_deltaV_deltaP".
	 * This is done by multiplying the reduced stoichiometric matrix Nr 
	 * of the given model with the matrix "deltaV_deltaP".
	 *
	 */
	private void init_Nr_times_DeltaV_DeltaP()
	{
		// now multiply Nr with deltaV_deltaP
		Nr_times_deltaV_deltaP = Nr.multiply(deltaV_deltaP);
	}

	/**
	 * Initiates the matrix "deltaV_deltaS_times_L".
	 * 	
	 */
	private void init_deltaV_deltaS_times_L()
	{
		Matrix deltaV_deltaS;
		Matrix L;
		// retrieve the elasticities at steady state
		double[][] dvals = jarnac.getUnScaledElasticityMatrix();  
		// now convert them to complex-values
		Complex[][] complex1 = new Complex[dvals.length][dvals[0].length];
		for (int i = 0; i < complex1.length; i++) 
			for (int ii = 0; ii < complex1[i].length; ii++) 
				complex1[i][ii] = Complex.valueOf(dvals[i][ii],0);
		// get the linker matrix
		dvals = model.getL().getArray();  
		Complex[][] complex2 = new Complex[dvals.length][dvals[0].length];
		for (int i = 0; i < complex2.length; i++) 
			for (int ii = 0; ii < complex2[i].length; ii++) 
				complex2[i][ii] = Complex.valueOf(dvals[i][ii],0);
		
		deltaV_deltaS = Matrix.valueOf(complex1);		
//		System.out.println("dv/ds");print(deltaV_deltaS);System.out.println("");System.out.println("");System.out.println("");		
		L = Matrix.valueOf(complex2);				
		deltaV_deltaS_times_L = deltaV_deltaS.multiply(L);		
	}

	private void init_jac_identity()
	{
		Complex[][] complex = new Complex[jacobian.getRowDimension()][jacobian.getColumnDimension()];
		for (int i = 0; i < complex.length; i++) 
			for (int ii = 0; ii < complex[i].length; ii++) 
				if (i==ii)
					complex[i][ii] = Complex.valueOf(1,0);
				else
					complex[i][ii] = Complex.valueOf(0,0);
		jac_identity = Matrix.valueOf(complex);		
	}
	
	/**
	 * Calculates the Frequency Response for a given model
	 *  
	 * 
	 * @param parameter 	the parameter for which the Frequency Response is calulated
	 * @param startPower  	the decimal power of the lowest frequency
	 * @param endPower		the decimal power of the highest frequency
	 * @param points		the number of points
	 */
	public void fResponse(String parameter, 			 
			int startPower,
			int endPower,
			int points)
	{		
		try 
		{
			Matrix temp;  // a working matrix;
			Matrix Hs;  // species concentration output;
			Matrix Hv;  // reaction rate output
			Matrix Csi;  // scaled concetration control coefficient
			Matrix Cv;  // scaled rate control coefficient
			Matrix Rsi;  // scaled concetration response coefficient
			Matrix Rv;  // scaled rate response coefficient
			
			double delta; // the delta between two measured frequencies
									
			Complex omega;   // the actual frequency
			LUDecomposition lu;   // LU Decomposition. Used to speed up the inversion of a matrix
			
			int counter = 0;
			int totalPointNumber;
			int pointsPerDecade;
			int decades;
					
			init_DeltaV_DeltaP(parameter);			
			init_deltaV_deltaS_times_L();
			init_Nr();
			init_Nr_times_DeltaV_DeltaP();
			init_jac_identity();			
								
			decades = (int)Math.abs(endPower - startPower);
			pointsPerDecade = Math.abs(Math.round(points / (endPower - startPower)));
			totalPointNumber = decades * pointsPerDecade;					
			
			spec_magnitude = new double[model.getISpecies().length][totalPointNumber];
			spec_phase = new double[model.getISpecies().length][totalPointNumber];
			
			rate_magnitude = new double[model.getReaction().length][totalPointNumber];
			rate_phase = new double[model.getReaction().length][totalPointNumber];

			frequency = new double[totalPointNumber];		
			for (int j = 0; j < decades; j++) 
			{
				for (int ii = 0; ii < pointsPerDecade; ii++) 
				{
					delta = 9.0 * (ii) / pointsPerDecade;						
					omega = Complex.valueOf(Math.pow(10, startPower+j) + 
											Math.pow(10, startPower+j) * delta, 0);
					
					frequency[counter] = omega.getReal();
					
					temp = (jac_identity.multiply(i.multiply(omega)).subtract(jacobian));
					lu = temp.lu();
					temp = lu.inverse();
					
					Hs = temp.multiply(Nr_times_deltaV_deltaP);
					Hv = deltaV_deltaS_times_L.multiply(Hs).add(deltaV_deltaP);					

					for (int iii = 0; iii < spec_magnitude.length; iii++) 
					{
						spec_magnitude[iii][counter] = 20 * Log10.log10(((Complex)Hs.get(iii,0)).magnitude());
						spec_phase[iii][counter] = Math.toDegrees(getArgument((Complex)Hs.get(iii,0)));
					}
					for (int iii = 0; iii < rate_magnitude.length; iii++) 
					{
						rate_magnitude[iii][counter] = 20 * Log10.log10(((Complex)Hv.get(iii,0)).magnitude());
						rate_phase[iii][counter] = Math.toDegrees(getArgument((Complex)Hv.get(iii,0)));					
					}	
					counter++;
				}
			}			
		}
		catch (NumberFormatException e) 
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		catch (Exception e) 
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * @return Returns the rate_gain.
	 */
	public double[][] getRate_magnitude() {
		return rate_magnitude;
	}
	/**
	 * @return Returns the rate_phase.
	 */
	public double[][] getRate_phase() {
		return rate_phase;
	}
	/**
	 * @return Returns the spec_gain.
	 */
	public double[][] getSpec_magnitude() {
		return spec_magnitude;
	}
	/**
	 * @return Returns the spec_phase.
	 */
	public double[][] getSpec_phase() {
		return spec_phase;
	}
	/**
	 * @return Returns the frequency.
	 */
	public double[] getFrequency() {
		return frequency;
	}
	
	private void showMessage(String message)
	{
		 JOptionPane.showMessageDialog(null, "Error occured", message, JOptionPane.ERROR_MESSAGE);
	}
	
	/**
	 * returns the argument for this complex number
	 * @param compl
	 * @return
	 */
	private double getArgument(Complex compl)
	{
		double z;
		z = compl.magnitude();
		if (compl.getImaginary() >= 0)
			return Math.acos(compl.getReal()/z);
		return -Math.acos(compl.getReal()/z);
	}		
}