/*
 * Created on 11.11.2004
 *
 */
package io.jni;

import math.ode.ODE;
import math.ode.ODE_Solver;


/**
 * @author Klaus Maier
 * 
 * contains the method which invokes lsoda which is written in C;
 * uses the Java Native Interface. 
 *
 */
public class LSODA extends ODE_Solver
{
	private ODE ode;
		
	/**
	 * evaluates the ode-system stored in the class variable "ode" for the given "y"-values
	 * at time "t".
	 * 
	 * this method is invoked from the native side, i.e. from the original lsoda-algorithm written 
	 * in C
	 * 
	 * @param y
	 * @param t
	 * @return
	 */
	private double[] eval(double[] y, double t)
	{	
		double[] result = new double[y.length];
		double[] yVals = new double[y.length-1];
		System.arraycopy(y, 1, yVals, 0, yVals.length);
		yVals = ode.evalODE(yVals,t);
				
		System.arraycopy(yVals, 0, result, 1, yVals.length);	
		return result;				
	}
	
	/**
	 * loads the lsoda library
	 */
    static 
	{
        System.loadLibrary("lsoda");
    }
    
	/**
	 * the native method which is used for solving odes.
	 * 
	 * the parameters are the same as used in the original lsoda-algorithm
	 * and therefore don't need to be explained here.
	 * 
	 * Further information about the parameters and their meanings can be found 
	 * in the source file lsoda.f in the internet at "netlib".
	 * 
	 * @param neq
	 * @param y
	 * @param t
	 * @param tout
	 * @param itol
	 * @param rtol
	 * @param atol
	 * @param itask
	 * @param istate
	 * @param iopt
	 * @param jt
	 * @param iwork1
	 * @param iwork2
	 * @param iwork5
	 * @param iwork6
	 * @param iwork7
	 * @param iwork8
	 * @param iwork9
	 * @param rwork1
	 * @param rwork5
	 * @param rwork6
	 * @param rwork7
	 */
	private native double[] lsoda(int neq, double[] y, double t, double tout, 
							      int itol, double[] rtol, double[] atol, int itask, 
							      int istate, int iopt, int jt, int iwork1, int iwork2, 
							      int iwork5, int iwork6, int iwork7, int iwork8, 
								  int iwork9, double rwork1, double rwork5, 
								  double rwork6, double rwork7); 
	
	
	/**
	 * inherited method from the abstract super-class ODE_Solver.
	 * uses the lsoda written in c for solving the given ode.
	 */
	public double[][] solve(ODE ode, double t, double[] timePoints, double[] yVals)	
	{
		double[][] result = new double[yVals.length][timePoints.length];		
		try
		{
			this.ode = ode;
			double[] y = new double[yVals.length + 1];	
		 	
	    	double rwork1, rwork5, rwork6, rwork7,tout;
	    	double[] atol = new double[yVals.length + 1];
	    	double[] rtol = new double[yVals.length + 1];
	    	int iwork1, iwork2, iwork5, iwork6, iwork7, iwork8, iwork9;
	    	int neq = yVals.length;
	    	int itol, itask, istate, iopt, jt, iout, j, jdum;
	
	    	iwork1= iwork2= iwork5= iwork6= iwork7= iwork8= iwork9= 0;
	    	rwork1= rwork5= rwork6= rwork7= 0.0;
	    	for (int i = 0; i < yVals.length; i++) 
	    	{
	    		rtol[i+1] = 1.0E-6;
	    		atol[i+1] = 1.0E-12;
				y[i+1] = yVals[i];
			}
	    	itol = 2;
	    	itask = 1;
	    	istate = 1;
	    	iopt = 0;
	    	jt = 2;    	
	    	
	    	if(timePoints[0]==t)
	    	{
				for (int i = 0; i < yVals.length; i++) 
				{
					result[i][0] = yVals[i];
				}
	    		
		    	for (int i = 1; i < timePoints.length; i++) 
		    	{
		    		tout = timePoints[i];
					y = lsoda(neq,y,t,tout,itol,rtol,atol,itask,istate,iopt,jt,
		    	    			iwork1, iwork2, iwork5, iwork6, iwork7, iwork8, iwork9,
								rwork1,rwork5,rwork6,rwork7);
					istate = 2;
					for (int ii = 1; ii < y.length; ii++) 
					{
						result[ii-1][i] = y[ii];
					}						
					t = timePoints[i];
				}
	    	}
	    	else
	    	{
		    	for (int i = 0; i < timePoints.length; i++) 
		    	{
		    		tout = timePoints[i];
					y = lsoda(neq,y,t,tout,itol,rtol,atol,itask,istate,iopt,jt,
		    	    			iwork1, iwork2, iwork5, iwork6, iwork7, iwork8, iwork9,
								rwork1,rwork5,rwork6,rwork7);
					istate = 2;
					for (int ii = 1; ii < y.length; ii++) 
					{
						result[ii-1][i] = y[ii];
					}						
					t = timePoints[i];
				}
	    		
	    	}
	    	
	    	
		}
		catch(Exception e)
		{
			System.out.println("Error on the native side");
		}
		
		return result;
	}	
	
}
