/*
 * Created on 02.11.2004
 *
 */
package model;

import io.java.BytecodeGenerator;
import io.java.EvalIF;
import io.java.JacobIF;

import java.io.IOException;
import java.util.LinkedList;
import java.util.Random;
import java.util.StringTokenizer;

import math.Echelon;
import math.JacobDiff;
import math.TMatrix;
import math.ode.ODE;
import math.sde.SDE;
import math.sto.STO;

import org.nfunk.jep.ParseException;

import Jama.Matrix;
import edu.caltech.sbw.*;
import sbw.Jarnac;
import sbw.NomInterface;


/**
 * @author Klaus Maier
 *	
 * A model is a description for some biological system.
 * In most cases models are created by parsing a sbml-file into the application 
 * 
 * The internal model representation used for this application relies for simplicity on the sbml-model definition
 * 
 * @TODO 
 * 1.FunctionDefinitions
 * 2.UnitDefinitions
 * 3.Rules
 * 4.Events
 * 
 */
public class Model implements SDE, STO, ODE
{
	//Avogadro's number is the number of particles in one mole of a substance.
	public final static double AVOGADRO = 6.022136736 * Math.pow( 10, 23);

	private static boolean haveNOM = false;
	// a unique identifier
	protected String name;
	// an array with processed formulas;
	// i.e. formulas for the internal representation
	// The Parameters are those stored in the allParameter-List.
	protected String[] javaReadableReactionRates;
	// Same formulas as in the above array. The only difference is that 
	// in this array the epression "two to the power of three" is stored as "2^3"
	// and in the former array it would have been Math.pow(2,3).
	// That is due to the fact that DJep (used for symbolically computing the Jacobian)
	// can work only with expressions of the former type
	protected String[] nonJavaReadableReactionRates;
	// this array contains the non-altered reaction rates
	protected String[] reactionRates;
	// as the name suggests this array contains all parameters; 
	// Internaly species with boundaryCondition=true are treated like a Parameter, i.e. this list can also 
	// contain Species-Objects
	protected LinkedList allParameter;
	
	protected Compartment[] compartment;
	// the "allSpecies" array contains all species, i.e. species with boundaryCondition = true and false
	protected Species[] allSpecies;
	// the "boundaryFalseSpecies" array contains species with boundaryCondition = false
	protected Species[] boundaryFalseSpecies;
	// the "boundaryTrueSpecies" array contains species with boundaryCondition = true
	protected Species[] boundaryTrueSpecies;
	
	// dependent species
	protected Species[] dSpecies;
	// independent species
	protected Species[] iSpecies;
	// in the "parameter" array are only the global parameters
	protected Parameter[] parameter;
	protected Reaction[] reaction;	
	
	// the full stoichiometric matrix
	protected TMatrix N;
	// the reduced stoichiometric matrix
	protected TMatrix Nr;
	// the linker matrix => N = L*Nr
	protected TMatrix L;
	// the part of L which is not part of the identity matrix
	protected TMatrix Lo;
	// the Matrix T is used to compute the values of independent variables
	// sd = Lo*si + T;
	protected TMatrix T;
		
	// used for generating random numbers when working with SDEs
	protected Random random;
	
	//contains the sbml as an String from which this model was created.
	protected String sbml;
	
	// the number of dependent variables; 
	protected int boundFalseSpecNum;
	// the number of independent variables;
	protected int boundTrueSpecNum;
	
	protected int parameterCount;	
//	private int functionIndex;	
	
	// the Jacobian for this model
	protected JacobIF jacobIF;
	
	// Objects implementing this interface are used to evaluat the right hand side of ODEs
	protected EvalIF evalIF;
	
	protected double[] specSteadyState; // the species concentrations at steady state
	protected double[] rateSteadyState; // the steady state rates
	
	protected double[] wiener;
	
	
	Module oNomModule = null;
	NomInterface oNom = null;
	String[] inBuiltFunctions = null;
	
	
	Jarnac jarnac;

	/**
	 * constructor
	 * 
	 * @param id
	 * @param name
	 * @param compartment
	 * @param species
	 * @param parameter
	 * @param reaction
	 */
	public Model( String name,
				  Compartment[] compartment,
				  Species[] species,
				  Parameter[] parameter,
				  Reaction[] reaction) throws Exception
	{
		this.name = name;
		
		this.compartment = compartment;
		this.allSpecies = species;
		this.parameter = parameter;
		this.reaction = reaction;
				
		this.init();
		// initiate the random number generator
		random = new Random(100);
	}
	/**
	 * constructor
	 * 
	 * @param id
	 * @param name
	 * @param compartment
	 * @param species
	 * @param parameter
	 * @param reaction
	 */
	public Model( String name,
				  Compartment[] compartment,
				  Species[] species,
				  Parameter[] parameter,
				  Reaction[] reaction,
				  String sbml) throws Exception
	{
		this(name,compartment,species,parameter,reaction);				
		this.sbml = sbml;
	}
	
	public Model (Model anotherModel) throws Exception
	{
		this(anotherModel.getName(),
				anotherModel.getCompartment(),
				anotherModel.getAllSpecies(),
				anotherModel.getParameter(),
				anotherModel.getReaction());
	}
	
	protected void init() throws Exception
	{		
		try
		{
			oNomModule = SBW.getModuleInstance("edu.caltech.NOMClipboard");
			oNom = (NomInterface) oNomModule.findServiceByName("NOM").getServiceObject(NomInterface.class);
			inBuiltFunctions = oNom.getBuiltinFunctions();
			haveNOM = true;
		}
		catch (SBWException e)
		{
			System.err.println ("Error communicating with NOM ... no support for inbuild functions.");
		}
		catch (Exception e)
		{
			System.err.println("No support for in build functions due to: " + e.getMessage());
		}
		// determine the number of dependent and independent variables for this model		
		boundFalseSpecNum = 0;
		boundTrueSpecNum = 0;
		for (int i = 0; i < allSpecies.length; i++) 
		{
			if (allSpecies[i].isBoundaryCondition()) 
			{
				boundTrueSpecNum++;				
			}
			else
			{
				boundFalseSpecNum++;
			}
		}
		
		// now determine and store the independent and dependent species in appropriate arrays
		boundaryFalseSpecies = new Species[this.boundFalseSpecNum];
		boundaryTrueSpecies = new Species[this.boundTrueSpecNum];
		int depCount = 0;
		int inDepCount = 0;
		for (int i = 0; i < allSpecies.length; i++) 
		{
			if (allSpecies[i].isBoundaryCondition()) 
			{
				this.boundaryTrueSpecies[inDepCount] = allSpecies[i];
				inDepCount++;
			}
			else
			{
				this.boundaryFalseSpecies[depCount] = allSpecies[i];
				depCount++;
			}
		}
				
		this.initODEs();
		this.initN();		
		
		Echelon echelon = new Echelon();
		this.Nr = echelon.getNr(N, 1E-12);
		this.L = echelon.getL(N, 1E-12);
		this.initLo();
				
		this.iSpecies = new Species[Nr.getRowDimension()];
		this.dSpecies = new Species[N.getRowDimension()-Nr.getRowDimension()];
		for (int i = 0; i < this.iSpecies.length; i++) 
		{
			this.iSpecies[i] = boundaryFalseSpecies[i];
		}
		for (int i = this.iSpecies.length; i < this.iSpecies.length + this.dSpecies.length; i++) 
		{
			this.dSpecies[i-this.iSpecies.length] = boundaryFalseSpecies[i];
		}
		
		this.initT();
		
		processReactionRates();

		String[] formulas = new String[boundaryFalseSpecies.length];
				
		for (int i = 0; i < N.getRowDimension(); i++)
		{			
			formulas[i] = "";
			for (int ii = 0; ii < N.getColumnDimension(); ii++) 
			{
					if(N.get(i,ii)!=0)
						if(formulas[i].length()>0)
							if(N.get(i,ii)<0)
								formulas[i] += N.get(i,ii)+"*"+ reaction[ii].getKineticLaw().getFormula();
							else
								formulas[i] += "+"+N.get(i,ii) +"*"+ reaction[ii].getKineticLaw().getFormula();
						else
							formulas[i] += N.get(i,ii) +"*"+ reaction[ii].getKineticLaw().getFormula();
			}			
		}

		try 
		{
			BytecodeGenerator.generateEvalClass( this, true );
		}
		catch (IOException e1) 
		{
			//System.out.println("Couldn't create an Evaluator class");			
			//e1.printStackTrace();
			throw new Exception("Could not compile model: " + e1.getMessage());
		}		

		try 
		{
			//this.initJacobian();
		}
		catch (Exception e)
		{
			System.err.println("Jacobian couldn't be created. This is not vital for CVODE ... error ignored.");						
		}
		
	}
	/**
	 * @return Returns the steadyState.
	 */
	public double[] getSpecSteadyState() {
		return specSteadyState;
	}
	/**
	 * @return Returns the steadyState.
	 */
	public double[] getRateSteadyState() {
		return rateSteadyState;
	}
	
//	public void calcSteadyState()
//	{
//		LSODA lsoda;
//		double[] vals;
//		
//		lsoda = new LSODA();	// initialize an ODE-solver
//		vals = new double[this.getBoundaryFalseSpecies().length];
//		
//		for (int i = 0; i < vals.length; i++) // get the initial concentrations and store them in vals
//			vals[i] = this.getBoundaryFalseSpecies()[i].getInitialConcentration(); 
//			
//		vals = lsoda.solve(this, 0, 100, vals);  // use the solver to simulate the system for 10 seconds.
//												// A good starting point is crucial for the Newton-Raphson-Method	
////		System.out.println(vals[0]+  "      "  +vals[1]);
//		try 
//		{
//			specSteadyState = NewtonRaphson.getSteadyState(100,vals,1E-12,1E-12,this);   // finally compute the steady state			
//			rateSteadyState = getReactionRates(specSteadyState);  // and the steady state rates		
//			
////			for (int i = 0; i < specSteadyState.length; i++) {
////				System.out.println(specSteadyState[i]);
////			}System.out.println("");
//		} 
//		catch (Exception e) 
//		{
//			System.out.println("Could not compute Steady State: "+e.getMessage());
////			e.printStackTrace();
//		}
//	}
	
//	public void calcSteadyState()
//	{
////		Jarnac jarnac = new Jarnac(sbml);
//		specSteadyState = jarnac.getSteadyState();
//		rateSteadyState = getReactionRates(specSteadyState);  
//	}
		
	/**
	 * @param jacobian The jacobian to set.
	 */
	public void setJacobian(JacobIF jacobian) {
		this.jacobIF = jacobian;
	}

	private String processFunction(String function, int reac, int functionIndex, String[] procJacForms)
	{
		StringTokenizer tok;
//		StringTokenizer closedBracketTokenizer;
		String token;
		String pow = "";
		String base = "";
		String func = "";
		String delim = "+-/*^()";
		String openBracket = "(";
		String closedBracket = ")";
		String power = "^";
		int index;
		int leftIndex;
		int rightIndex;
		int oBracketCounter;
		int cBracketCounter;
		tok = new StringTokenizer(function, delim, true);
		while(tok.hasMoreTokens())
		{				
			token = tok.nextToken().toLowerCase();
			functionIndex += token.length();
			
			if(delim.indexOf(token) >= 0) // if true => token equals +,-,*,/,(,) or ^
				if(!token.equals(power))
				{
					func += token;
					procJacForms[reac] += token;
				}
				else // token = "^"
				{		
					procJacForms[reac] += token;
					leftIndex = func.length();
					
					// check whether the base is a single number or variable or whether it is 
					// a whole expression
					if (func.substring(func.length()-1, func.length()).equals(closedBracket)) // if true base is a expresion (...)
					{			
						oBracketCounter = 0;						
						for (int i = func.length()-1; i > -1; i--) 
						{
							if (func.substring(i, i+1).equals(closedBracket))
								oBracketCounter++;
							else
								if (func.substring(i, i+1).equals(openBracket))
									oBracketCounter--;
							leftIndex--;								
							if (oBracketCounter == 0)
								break;
						}						
					}
					else // base is only a single variable or number
					{
						for (int i = func.length()-1; i > -1; i--) // get all literals unless it is an operator (+,-,/,*)
						{	
							if(delim.indexOf(func.substring(i, i+1)) >= 0)
								break;
							else
								leftIndex--;
							
						}
					}
					
					// now get the power
					// test if pow is a single variable/number or whether it is
					// a whole expression (...)
					if (function.substring(functionIndex, functionIndex+1).equals(openBracket)) // true if next literal is a bracket "("
					{						
						oBracketCounter = 0;	
						cBracketCounter = 0;
						for (int i = functionIndex; i < function.length(); i++) 
						{
							if (function.substring(i, i+1).equals(closedBracket))
							{
								oBracketCounter--;
								cBracketCounter++;
							}
							else
								if (function.substring(i, i+1).equals(openBracket))
									oBracketCounter++;							
							if (oBracketCounter == 0)
								break;
						}						
						pow = "";
						while(cBracketCounter>0)
						{
							token = tok.nextToken().toLowerCase();
							functionIndex += token.length();
							pow += token;							
							if (token.equals(closedBracket))
								cBracketCounter--;
						}
						pow = processFunction (pow, reac, functionIndex, procJacForms);
					}
					else // next literal wasn't a bracket;  
						 // must be a variable or number
					{
						token = tok.nextToken().toLowerCase();
						functionIndex += token.length();
						pow = processFunction (token, reac, functionIndex, procJacForms);
					}
					
					
					base = func.substring(leftIndex, func.length());
					// remove the old tokens
					func = func.substring(0, leftIndex);
					
					func += "Math.pow("+base+","+pow+")";					
				}
			else // token is a Parameter Species or a number
			{
				index = -1;
				if (reaction[reac].getKineticLaw().getLocalParameter()!=null)
					for (int ii = 0; ii < reaction[reac].getKineticLaw().getLocalParameter().length; ii++) 
					{
						if (token.equals(reaction[reac].getKineticLaw().getLocalParameter()[ii].getName().toLowerCase()))
						{
							index = ii;
							break;
						}
					}
				if(index!=-1) // it's a local Parameter
				{
					func += "p"+parameterCount;
					procJacForms[reac] += "p"+parameterCount;
					parameterCount++;
					allParameter.add(reaction[reac].getKineticLaw().getLocalParameter()[index]);
				}
				else
				{
					for (int ii = 0; ii < boundaryFalseSpecies.length; ii++) 
					{
						if (token.equals(boundaryFalseSpecies[ii].getName().toLowerCase()))
						{
							index = ii;
							break;
						}
					}
					if(index!=-1) // it's a species
					{
						func += "y["+index+"]";
						procJacForms[reac] += "y"+index;
					}
					else // so let's check whether it is a global Parameter 
					{
						for (int ii = 0; ii < boundaryTrueSpecies.length; ii++) 
						{
							if (token.equals(boundaryTrueSpecies[ii].getName().toLowerCase()))
							{
								index = ii;
								break;
							}
						}
						if(index!=-1) // it's an boundary=true species
						{
							func += "p"+parameterCount;
							procJacForms[reac] += "p"+parameterCount;
							parameterCount++;
							allParameter.add(new Parameter(boundaryTrueSpecies[index].getName(), 
									boundaryTrueSpecies[index].getUnits(), 
									boundaryTrueSpecies[index].getInitialConcentration()));
						}
						else
						{
							if (parameter!=null)
							{
								for (int ii = 0; ii < parameter.length; ii++) 
								{
									if (token.equals(parameter[ii].getName().toLowerCase()))
									{
										index = ii;
										break;
									}
								}
								if(index!=-1) // it's a local Parameter
								{
									func += "p"+parameterCount;
									procJacForms[reac] += "p"+parameterCount;
									parameterCount++;
									allParameter.add(parameter[index]);
								}
							}
							if(index == -1) // still not found
							{
								if (token.equals("log")
										|| token.equals("abs")
										|| token.equals("acos")
										|| token.equals("asin")
										|| token.equals("atan")
										|| token.equals("ceil")
										|| token.equals("cos")
										|| token.equals("exp")
										|| token.equals("floor")
										|| token.equals("sin")
										|| token.equals("sqrt")
										|| token.equals("tan") )
									
								{									
									func += "Math." + token;
									procJacForms[reac] += token;
								}
								else if (token.equals("ln") || token.equals("log10"))
								{
									func += "Math.log";
									procJacForms[reac] += token;
								}
								else if (token.equals("pow"))
								{
									procJacForms[reac] += token;
									func += "Math.pow";															
								}							
								else if (isInBuildFunction(token) && oNom != null && haveNOM)
								{
									try
									{
										String[] functionDefinition = oNom.getBuiltinFunctionInfo(token);
										int nNumParameters = functionDefinition.length - 3;
										String sFunction = functionDefinition[functionDefinition.length - 1];
										String[] functionParameter = new String[nNumParameters];
										System.arraycopy(functionDefinition,2,functionParameter,0,nNumParameters);
										
										// read opening bracket
										token = tok.nextToken();
										func += token;
										functionIndex++;
										
										int nLength = 0;
										
										oBracketCounter = 0;	
										cBracketCounter = 0;
										
										for (int i = functionIndex -1; i < function.length(); i++) 
										{
											if (function.substring(i, i+1).equals(closedBracket))
											{
												oBracketCounter--;
												cBracketCounter++;
											}
											else if (function.substring(i, i+1).equals(openBracket))
											{
													oBracketCounter++;
											}
											if (oBracketCounter == 0)
												break;
										}						
										pow = "";
										while(cBracketCounter>0)
										{
											token = tok.nextToken().toLowerCase();
											functionIndex += token.length();
											pow += token;							
											if (token.equals(closedBracket))
												cBracketCounter--;
										}
							
										pow = pow.substring(0,pow.length()-1);
										String oParams[] = pow.split(",");										
										if (oParams.length != functionParameter.length)
											throw new IllegalArgumentException("Mismatch while parsing inbuilt function.");
										for (int k = 0; k < functionParameter.length; k++)
										{
											String sTemp = processFunction (oParams[k], reac, functionIndex, procJacForms);
											if (sTemp.charAt(0) != '(' || sTemp.charAt(sTemp.length() -1) != ')')
												sTemp = "(" + sTemp + ")";
											sFunction = sFunction.replaceAll(functionParameter[k], sTemp); 
										}
										func += sFunction + ")";
										
										// read closing bracket
										token = tok.nextToken();
										func += token;
										functionIndex++;
									}
									catch (Exception e)
									{
										// ignore
									}
								}
								else // must be a number    <! --- this is a lie :)
								{					
									if (token.indexOf(",") != -1)	//<! -- could be some arguments
									{
										String[] parameters = token.split(",");
										for (int i = 0; i < parameters.length; i ++)
										{
											func += processFunction (parameters[i], reac, functionIndex, procJacForms);
											if (i+1 < parameters.length)
											{
												func += ",";
											}
										}
									}
									else
									{
										// test whether parameter is a number ... if so convert it to double. 
										func += new Double(token).toString();	
										procJacForms[reac] += new Double(token).toString();;
									}
								}
							}
						}
					}
				}
			}
		}
		return func;
	}
	
	private boolean isInBuildFunction (String sName)
	{
		for (int i = 0; i < inBuiltFunctions.length; i++)
		{
			if (inBuiltFunctions[i].equals(sName))
				return true;
		}
		return false;
	}
	public void initJacobian() throws Exception
	{
		try 
		{
			JacobDiff jacobDiff = new JacobDiff(this);
			try 
			{
				BytecodeGenerator.generateJacobClass(this, true, jacobDiff.getDerivatives());
			} 
			catch (IOException e1) 
			{
				throw new IOException("Could not create Jacobian: " + e1.getMessage());
			}
		} 
		catch (ParseException e) 
		{
			throw new ParseException("Could not create Jacobian: " + e.getMessage());
		}
	}
	
	public String[] get_jepReadableReactionRates() {
		return nonJavaReadableReactionRates;
	}
	public String[] get_reactionRates() {
		return reactionRates;
	}
	/**
	 * @return Returns the nr.
	 */
	public TMatrix getNr() {
		return Nr;
	}
	/**
	 * @return Returns the l.
	 */
	public TMatrix getL() {
		return L;
	}
	/**
	 * @return Returns the lo.
	 */
	public TMatrix getLo() {
		return Lo;
	}
	/**
	 * @return Returns the t.
	 */
	public TMatrix getT() {
		return T;
	}	
	/**
	 * @return Returns the n.
	 */
	public TMatrix getN() {
		return N;
	}

	/**
	 * this method is used to process the functions stored in the KineticLaw-Objects;
	 * Species (i.e. their IDs) will be represented with s[i] to s[x];
	 * Parameters (i.e. their IDs) will be represented with p1 to px; Species
	 * which have boundaryCondition = true will be treated like Parameters; 
	 * 
	 */
	private void processReactionRates()
	{
		allParameter = new LinkedList();
		javaReadableReactionRates = new String[reaction.length];
		nonJavaReadableReactionRates = new String[reaction.length]; 
		reactionRates = new String[reaction.length]; 
		parameterCount = 0;
		for (int i = 0; i < javaReadableReactionRates.length; i++) 
		{
			nonJavaReadableReactionRates[i] = "";
			reactionRates[i] = reaction[i].getKineticLaw().getFormula();
			javaReadableReactionRates[i] = processFunction(reaction[i].getKineticLaw().getFormula(),i,0,nonJavaReadableReactionRates);	
		}
	}
	
	/**
	 * @return Returns the allParameter.
	 */
	public LinkedList getAllParameter() {
		return allParameter;
	}
	
	public String[] get_javaReadableReactionRates() {
		return javaReadableReactionRates;
	}
	/**
	 * @return Returns the id.
	 */
	public String getName() {
		return name;
	}
	/**
	 * @return Returns the compartment.
	 */
	public Compartment[] getCompartment() {
		return compartment;
	}
	/**
	 * @return Returns the parameter.
	 */
	public Parameter[] getParameter() {
		return parameter;
	}
	/**
	 * @return Returns the reaction.
	 */
	public Reaction[] getReaction() {
		return reaction;
	}
	/**
	 * @return Returns the sbml.
	 */
	public String getSbml() {
		return sbml;
	}
	/**
	 * @param sbml The sbml to set.
	 */
	public void setSbml(String sbml) {
		this.sbml = sbml;
	}
	
	/**
	 * inherited from math.ode.ODE;
	 * evaluates this model for the given values. 
	 * returns an array with the results. The results are concetrations.
	 * 
	 */
	public double[] evalODE(double[] yVals, double x) 
	{		
		return evalIF.eval(yVals, x);
	}
	/**
	 * @return Returns the allSpecies.
	 */
	public Species[] getAllSpecies() {
		return allSpecies;
	}
	/**
	 * @return Returns the iSpecies.
	 */
	public Species[] getISpecies() {
		return iSpecies;
	}
	/**
	 * @return Returns the dSpecies.
	 */
	public Species[] getDSpecies() {
		return dSpecies;
	}
	
	public double[] getDependentSpeciesValues(double[] iSpeciesVals)
	{
		double[] sd;
		Matrix si = new Matrix(iSpeciesVals, iSpeciesVals.length);		
		sd = (Lo.times(si).plus(T)).getRowPackedCopy();		
		return sd;
	}
	
	public double[] evalIndependent(double[] yVals)
	{
		double[] rightSide = new double[iSpecies.length];
		double[] rates = new double[reaction.length];
		for (int i = 0; i < rates.length; i++) 
		{
			rates[i] = this.evalIF.getReactionRate(i, yVals);
		}
		for (int i = 0; i < Nr.getRowDimension(); i++) 
		{
			for (int ii = 0; ii < Nr.getColumnDimension(); ii++) 
			{
				if (Nr.get(i,ii)!=0)
					rightSide[i] += Nr.get(i,ii) * rates[ii];
			}
		}
		return rightSide;
	}
	
	
	
	/**
	 * @param evaluator The evaluator to set.
	 */
	public void setEvaluator(EvalIF evaluator) {
		this.evalIF = evaluator;
	}
	
	/**
	 * @return Returns the evaluator.
	 */
	public EvalIF getEvaluator() {
		return evalIF;
	}

	/**
	 * returns the number of boundary=false variables for the underlying ODE-System
	 */
	public int getBoundFalseSpecNum() 
	{
		return boundFalseSpecNum;
	}
	/**
	 * returns the number of boundary=true variables for the underlying ODE-System
	 */
	public int getBoundTrueSpecNum() 
	{
		return boundTrueSpecNum;
	}
	/**
	 * returns the number of dependent variables for the underlying ODE-System
	 */
	public int getDepSpecNum() 
	{
		return this.dSpecies.length;
	}
	/**
	 * returns the number of dependent variables for the underlying ODE-System
	 */
	public int getIndSpecNum() 
	{
		return this.iSpecies.length;
	}	
	
	/**
	 * returns the Comparable out of the given array whose identifier equals "id"
	 * 
	 * @param array
	 * @param id
	 * @return
	 */
	private Comparable getComparable (Comparable[] array, String id)
	{
		boolean found = false;
		int count = 0;
		while (!found && 
			   count < array.length)
		{
			if(array[count].equals(id))
				return array[count];
			else
				count++;
		}
		return null;
	}
	
	/**
	 * used at the very beginning to initialize and create the ODEs for this model
	 *
	 */
	public void initODEs()
	{
		try 
		{			
			for (int i = 0; i < allSpecies.length; i++) 
			{
				Compartment comp = ((Compartment)getComparable(compartment, allSpecies[i].getCompartment()));	
				allSpecies[i].setInitialConcentration( allSpecies[i].getInitialAmount() / comp.getVolume() );		
				allSpecies[i].setComp(((Compartment)getComparable(compartment, allSpecies[i].getCompartment())));				
			}
		} 
		catch (Exception e) 
		{
			e.printStackTrace();
		}		
	}
		
	public double[] evalSDE(double[] vals, double x) 
	{		
		return evalODE(vals, x);		
	}	
	/**
	 * inherited from math.ode.ODE;
	 * evaluates this model for the given values. 
	 * returns an array with the results. The results are concetrations.
	 * 
	 */
	public double[] evalNoise(double[] yVals, double x) 
	{		
		double result[] = new double[getBoundFalseSpecNum()];
		
		wiener = new double[getReactNum()];
		for (int i = 0; i < wiener.length; i++) 
			wiener[i] = random.nextGaussian();		
		
		for (int i = 0; i < getBoundFalseSpecNum(); i++)
		{
			result[i] = this.evalNoise(yVals,i);
			
		}		
		return result;
	}
	private double evalNoise(double[] y, int spec) 
	{
		double res = 0;			

		for (int equat = 0; equat < N.getColumnDimension(); equat++) 
		{
			if (N.get(spec, equat)!= 0) 
			{
				res += N.get(spec, equat) * Math.sqrt(this.evalIF.getReactionRate(equat)) * wiener[equat];
			}
		}		
		res /= boundaryFalseSpecies[spec].getComp().getVolume();		
		return res;		
	}
	
	private void initT()
	{
		double tem;
		this.T = new TMatrix(this.N.getRowDimension() - this.Nr.getRowDimension(),1);
		double[] iVals = new double[this.iSpecies.length];
		for (int i = 0; i < iVals.length; i++) 
		{
			iVals[i] = this.iSpecies[i].getInitialConcentration();
		}
		Matrix iVec = new Matrix(iVals, iVals.length);
				
		Matrix temMat = Lo.times(iVec);
		
		for (int i = 0; i < T.getRowDimension(); i++) 
		{
			T.set(i,0,this.dSpecies[i].getInitialConcentration() - temMat.get(i,0));
		}
	}
	
	private void initLo()
	{
		this.Lo = new TMatrix(this.L.getRowDimension()-this.Nr.getRowDimension(), this.L.getColumnDimension());
		for (int i = this.Nr.getRowDimension(); i < this.L.getRowDimension(); i++) 
		{
			for (int ii = 0; ii < this.L.getColumnDimension(); ii++) 
			{
				this.Lo.set(i-this.Nr.getRowDimension(),ii,this.L.get(i,ii));
			}
		}
	}
	
	/**
	 * used to initiate the stoichiometry matrix for this model
	 *
	 */
	private void initN()
	{
		double[][] n = new double[this.boundFalseSpecNum][reaction.length];
		int index;
		for (int i = 0; i < reaction.length; i++) 
		{
			for (int ii = 0; ii < reaction[i].getProducts().length; ii++) 
			{
				index = -1;
				for (int iii = 0; iii < this.boundaryFalseSpecies.length; iii++) 
				{
					if (reaction[i].getProducts()[ii].getName().equals(boundaryFalseSpecies[iii].getName())) 
					{
						index = iii;
						break;
					}
				}
				if(index!=-1)
				{
					n[index][i] += reaction[i].getProducts()[ii].getStoichiometry();
				}
			}
			for (int ii = 0; ii < reaction[i].getReactants().length; ii++) 
			{
				index = -1;
				for (int iii = 0; iii < this.boundaryFalseSpecies.length; iii++) 
				{
					if (reaction[i].getReactants()[ii].getName().equals(boundaryFalseSpecies[iii].getName())) 
					{
						index = iii;
						break;
					}
				}
				if(index!=-1)
				{
					n[index][i] -= reaction[i].getReactants()[ii].getStoichiometry();
				}			
			}
		}
		
		this.N = new TMatrix(n, this);
	}
	
	/**
	 * @return Returns the depSpecies.
	 */
	public Species[] getBoundaryFalseSpecies() {
		return boundaryFalseSpecies;
	}
	/**
	 * @return Returns the inDepSpecies.
	 */
	public Species[] getBoundaryTrueSpecies() {
		return boundaryTrueSpecies;
	}
	/**
	 * returns the number of reactions in this model. 
	 * This method is used by the stochastic algorithms.
	 *  
	 * @see math.sto.STO#getReactNum()
	 */
	public int getReactNum() 
	{		
		return reaction.length;
	}
	/**
	 * 
	 * @see math.sto.STO#execReaction(int, int[])
	 */
	public double[] execReaction(int reac, double[] molecules) 
	{		
//		int counter = -1;
		for (int i = 0; i < N.getRowDimension(); i++) 
		{
			molecules[i] += N.get(i,reac);
			if (molecules[i]<0)
				molecules[i] = 0;
		}		
		return molecules;		
	}
	/* (non-Javadoc)
	 * @see math.stoch.STO#getPropensity(int, int[])
	 */
	public double getPropensity(int reac, double[] molecules) 
	{
		return this.evalIF.getPropensity(reac, molecules);
	}

	public Matrix getJacobian(double[] vals)
	{
		Matrix elas = new Matrix(jacobIF.get_SpeciesElasticityCoefficients(vals));
		return (Nr.times(elas)).times(L);
	}
	public double[][] getJacobianArray(double[] vals)
	{
		return getJacobian(vals).getArray();
	}
	public double[] getReactionRates(double[] vals)
	{
		double[] rates = new double[reaction.length];
		for (int i = 0; i < rates.length; i++) 
			rates[i] = this.evalIF.getReactionRate(i,vals);
		return rates;
	}
	/* (non-Javadoc)
	 * @see math.langevin.SDE#getJacobian()
	 */
	public JacobIF getJacobIF() 
	{
		return jacobIF;
	}
	
	public void updateReactionChange(int reac1, int reac2)
	{
		Reaction temp = this.reaction[reac1].getClone();
		this.reaction[reac1] = this.reaction[reac2];
		this.reaction[reac2] = temp;
	}
	public void updateSpeciesChange(int spec1, int spec2)
	{
		Species temp = this.boundaryFalseSpecies[spec1].getClone();
		this.boundaryFalseSpecies[spec1] = this.boundaryFalseSpecies[spec2];
		this.boundaryFalseSpecies[spec2] = temp;
	}
	
	public void initJarnac(String sbml) throws SBWException
	{
		if (Sys.OSIsWindows())
		{
		this.jarnac = new Jarnac();
		this.jarnac.loadSBML(sbml);
		}
	}
	
	public Jarnac getJarnac()
	{
		if (Sys.OSIsWindows())
		{
		return jarnac;
		}
		return null;
	}
}