/*
 * Created on 09.10.2004
 *
 */
package sbml;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.StringReader;
import java.io.IOException;
import java.util.List;

import model.Compartment;
import model.KineticLaw;
import model.Model;
import model.Parameter;
import model.Reaction;
import model.Species;
import model.SpeciesReference;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;

import edu.caltech.sbw.*;
import sbw.NomInterface;

/**
 * @author Klaus Maier
 *
 * SBML_Reader objects are used to load sbml-files into the application.
 * Uses the free available jdom-package. Information about jdom can be found
 * in the internet.
 * 
 * 
 * @TODO 
 * 1.FunctionDefinitions
 * 2.UnitDefinitions
 * 3.Rules
 * 4.Events
 * 5.Level1 sbml -> Reaction: stoichiometry can be divided into numerator and denominator
 *  
 */
public class SBML_Reader 
{		
	

	public static String getSBML(File aFile) throws IOException
	{		
		String sbml = "";
		FileReader fileReader;
		BufferedReader bufferedReader;
		fileReader = new FileReader(aFile);
		bufferedReader = new BufferedReader(fileReader);
		String line;
		line = bufferedReader.readLine(); 
	    while ( line != null ) 
	    { 		        
	    	sbml += line + "\n";
	    	line = bufferedReader.readLine(); 	    	
	    } 
		bufferedReader.close();
		return sbml;
	}

	
	/**
	 * receives a sbml-file and returns the corresponding model
	 * 
	 * @param aFile
	 * @return
	 * @throws Exception
	 */
	public static Model load(File aFile) throws Exception
	{
		return load(new FileReader(aFile));
	}
	static Model load(java.io.Reader oReader) throws Exception
	{
		String name;
		
		Model model;
		
		Compartment[] compartments;
		Species[] species;
		Reaction[] reactions;
		Parameter[] parameters;	
		
		String sbml;
		
		List listOfFunctionDefinitions = null;
		List listOfUnitDefinitions = null;
		List listOfCompartments = null;
		List listOfSpecies = null;
		List listOfParameters = null;
		List listOfRules = null;
		List listOfReactions = null;
		
		SAXBuilder builder = new SAXBuilder();
		
		try 
		{			
			Document doc = builder.build( oReader );			
			Element root = doc.getRootElement();
			Element child;			
						
			Namespace ns = root.getNamespace();	
			
			// get the root (model) and its values
			child = root.getChild( SBML_Identifiers.MODEL, ns );		
			
//			List lis = root.getChildren();
//			child = ((Element)lis.get(0));
			
			name = child.getAttributeValue( SBML_Identifiers.NAME);			

			// now check if level 1; 
			// in the moment only level 1 is supported			
			if (!isLevel1(ns))
			{
				throw new IllegalArgumentException("Currently only SBML level 1 is supported."); 
			}
									
			// get the lists
			if (child.getChild(SBML_Identifiers.listOfFunctionDefinitions, ns) != null)
				listOfFunctionDefinitions = child.getChild(SBML_Identifiers.listOfFunctionDefinitions, ns).getChildren();
			if (child.getChild(SBML_Identifiers.listOfUnitDefinitions, ns) != null)
				listOfUnitDefinitions = child.getChild(SBML_Identifiers.listOfUnitDefinitions, ns).getChildren();
			if (child.getChild(SBML_Identifiers.listOfCompartments, ns) != null)
				listOfCompartments = child.getChild(SBML_Identifiers.listOfCompartments, ns).getChildren();
			if (child.getChild(SBML_Identifiers.listOfSpecies, ns) != null)
				listOfSpecies = child.getChild(SBML_Identifiers.listOfSpecies, ns).getChildren();
			if (child.getChild(SBML_Identifiers.listOfParameters, ns) != null)
				listOfParameters = child.getChild(SBML_Identifiers.listOfParameters, ns).getChildren();
			if (child.getChild(SBML_Identifiers.listOfRules, ns) != null)	
				listOfRules = child.getChild(SBML_Identifiers.listOfRules, ns).getChildren();
			if (child.getChild(SBML_Identifiers.listOfReactions, ns) != null)
				listOfReactions = child.getChild(SBML_Identifiers.listOfReactions, ns).getChildren();
			
			// get the compartments;
			compartments = extractCompartments (listOfCompartments) ;
			// get the Species
			species = extractSpecies (listOfSpecies) ;
			// get the global parameters
			parameters = extractParameters (listOfParameters) ;
			// get the Reactions
			reactions = extractReactions (listOfReactions) ;
			
			model = new Model ( name,
								compartments,
								species,
								parameters,
								reactions);	
					
			return model;
		}		 
		catch (JDOMException e) 
		{
			// TODO Auto-generated catch block
			throw new IllegalArgumentException("Couldn't parse SBML file: " + e.getMessage());
		} 
		catch (IOException e) 
		{
			throw new IllegalArgumentException("Couldn't read SBML file: " + e.getMessage());
			// TODO Auto-generated catch block
		}
	
	}
	public static Model load(String sbml) throws Exception
	{		
		String sModel = sbml;
		try
		{
			Module oNomModule = SBW.getModuleInstance("edu.caltech.NOMClipboard");
			NomInterface oNom = (NomInterface) oNomModule.findServiceByName("NOM").getServiceObject(NomInterface.class);
			sModel = oNom.convertLevel2ToLevel1Impl(sbml);
		}
		catch (Exception e)
		{
			System.err.println("Couldn't find NOM ... Can load only SBML level 1 models.");
		}
		
		return load(new StringReader(sModel));
	}
	
	/**
	 * returns true if the given namespace is SBML-Level1, else false
	 * @param ns
	 * @return
	 */
	private static boolean isLevel1(Namespace ns)
	{
		if(ns.equals(Namespace.getNamespace("http://www.sbml.org/sbml/level1")))
			return true;
		return false;
	}


	/**
	 * extract the compartments from the list and return them in the return-array;
	 * 
	 * @param list
	 * @return
	 */
	private static Reaction[] extractReactions(List list) throws Exception
	{
		if (list == null)
			return null;

		String name;
		String specName;
		
		SpeciesReference[] reactants = null;
		SpeciesReference[] products = null;
		
		double stoi;
		
		boolean reversible;
		boolean fast;
		boolean found;
		
		// list with references to species used as reactants,...
		List listOfReactants = null;
		List listOfProducts = null;
		
		Namespace ns = ((Element)list.get(0)).getNamespace();
		
		Reaction[] reaction = new Reaction[list.size()];
		
		KineticLaw kineticLaw = null;		
		Element sbmlKineticLaw;		
		
		for (int i = 0; i < reaction.length; i++) 
		{
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.NAME) != null)
				name = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.NAME);
			else
				name = "";						
						
			// get the lists
			if (((Element)list.get(i)).getChild(SBML_Identifiers.listOfReactants, ns) != null)
				listOfReactants = ((Element)list.get(i)).getChild(SBML_Identifiers.listOfReactants, ns).getChildren();
			else
				listOfReactants = null;
			if (((Element)list.get(i)).getChild(SBML_Identifiers.listOfProducts, ns) != null)
				listOfProducts = ((Element)list.get(i)).getChild(SBML_Identifiers.listOfProducts, ns).getChildren();
			else
				listOfProducts = null;
			
			// now extract the information from the lists
			if (listOfReactants != null)
			{
				reactants = new SpeciesReference[listOfReactants.size()];
				for (int ii = 0; ii < listOfReactants.size(); ii++) 
				{				
					if (((Element)listOfReactants.get(ii)).getAttributeValue(SBML_Identifiers.SPECIES)!=null)
						specName = ((Element)listOfReactants.get(ii)).getAttributeValue(SBML_Identifiers.SPECIES);
					else
						specName = ((Element)listOfReactants.get(ii)).getAttributeValue(SBML_Identifiers.SPECIE);
					
					if ( ((Element)listOfReactants.get(ii)).getAttributeValue(SBML_Identifiers.STOICHIOMETRY) != null )
						stoi = Double.parseDouble( ((Element)listOfReactants.get(ii)).getAttributeValue(SBML_Identifiers.STOICHIOMETRY) );
					else
						stoi = 1;
					if (( ((Element)listOfReactants.get(ii)).getAttributeValue(SBML_Identifiers.DENOMINATOR) != null ))
						stoi /= Double.parseDouble( ((Element)listOfReactants.get(ii)).getAttributeValue(SBML_Identifiers.DENOMINATOR) );

					reactants[ii] = new SpeciesReference(specName, stoi);
				}
			}
			if (listOfProducts != null)
			{
				products = new SpeciesReference[listOfProducts.size()];
				for (int ii = 0; ii < listOfProducts.size(); ii++) 
				{				
					if (((Element)listOfProducts.get(ii)).getAttributeValue(SBML_Identifiers.SPECIES)!=null)
						specName = ((Element)listOfProducts.get(ii)).getAttributeValue(SBML_Identifiers.SPECIES);
					else
						specName = ((Element)listOfProducts.get(ii)).getAttributeValue(SBML_Identifiers.SPECIE);
					
					if ( ((Element)listOfProducts.get(ii)).getAttributeValue(SBML_Identifiers.STOICHIOMETRY) != null )
						stoi = Double.parseDouble( ((Element)listOfProducts.get(ii)).getAttributeValue(SBML_Identifiers.STOICHIOMETRY) );
					else
						stoi = 1;
					if (( ((Element)listOfProducts.get(ii)).getAttributeValue(SBML_Identifiers.DENOMINATOR) != null ))
						stoi /= Double.parseDouble( ((Element)listOfProducts.get(ii)).getAttributeValue(SBML_Identifiers.DENOMINATOR) );

					products[ii] = new SpeciesReference(specName, stoi);
				}	
			}				
				
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.REVERSIBLE) != null)
				if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.REVERSIBLE).equals("true"))
					reversible = true;
				else
					reversible = false;
			else
				reversible = true;	
			
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.FAST) != null)
				if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.FAST).equals("true"))
					fast = true;
				else
					fast = false;				
			else
				fast = false;	
			
			if (((Element)list.get(i)).getChild(SBML_Identifiers.KINETIC_LAW, ns) != null)
			{
				sbmlKineticLaw = ((Element)list.get(i)).getChild(SBML_Identifiers.KINETIC_LAW, ns);
				
				String timeUnits = "";
				String substanceUnits = "";
				String formula = "";
				
				Parameter[] localParameters = null;
								
				if (sbmlKineticLaw.getAttributeValue(SBML_Identifiers.TIME_UNITS) != null)
					timeUnits = sbmlKineticLaw.getAttributeValue(SBML_Identifiers.TIME_UNITS);
				if (sbmlKineticLaw.getAttributeValue(SBML_Identifiers.SUBSTANCE_UNITS) != null)
					substanceUnits = sbmlKineticLaw.getAttributeValue(SBML_Identifiers.SUBSTANCE_UNITS);
				
				if (sbmlKineticLaw.getAttributeValue(SBML_Identifiers.FORMULA) != null)
					formula = sbmlKineticLaw.getAttributeValue(SBML_Identifiers.FORMULA);	
				
				if (sbmlKineticLaw.getChild(SBML_Identifiers.listOfParameters, ns) != null)
					localParameters = extractParameters ( sbmlKineticLaw.getChild(SBML_Identifiers.listOfParameters, ns).getChildren() );
								
				kineticLaw = new KineticLaw (timeUnits,
										     substanceUnits,
											 localParameters,
											 formula);				
			}		
			
			reaction[i] = new Reaction( name,
										reactants,
										products,
										reversible,
										fast,
										kineticLaw);
		}			
		return reaction;
	}	
	
	/**
	 * extract the compartments from the list and return them in the return-array
	 * 
	 * @param list
	 * @return
	 */
	private static Compartment[] extractCompartments(List list) throws Exception
	{
		if (list == null)
			return null;
		
		String name;
		String outside;
		String units;
						
		double volume;
		
		Compartment[] compartment = new Compartment[list.size()];
		
		for (int i = 0; i < compartment.length; i++) 
		{
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.NAME) != null)
				name = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.NAME);
			else
				name = "";		
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.UNITS) != null)
				units = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.UNITS);
			else
				units = "";		
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.OUTSIDE) != null)
				outside = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.OUTSIDE);
			else
				outside = "";
			
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.VOLUME) != null)
				volume = Double.parseDouble( ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.VOLUME));
			else
				volume = Double.NaN;		
						
			compartment[i] = new Compartment( name,
											  units,
											  outside,
											  volume);
		}			
		return compartment;
	}	
	
	/**
	 * extract the parameters from the list and return them in an array
	 * 
	 * @param list
	 * @return
	 */
	private static Parameter[] extractParameters(List list) throws Exception
	{
		if (list == null)
			return null;
		
		String name;
		String units;
		
		double value;
				
		Parameter[] parameter = new Parameter[list.size()];
				
		for (int i = 0; i < parameter.length; i++) 
		{			
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.NAME) != null)
				name = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.NAME);
			else
				name = "";				
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.UNITS) != null)
				units = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.UNITS);
			else
				units = "";			

			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.VALUE) != null)
				value = Double.parseDouble(((Element)list.get(i)).getAttributeValue(SBML_Identifiers.VALUE));
			else
				value = Double.MIN_VALUE;		
											
			parameter[i] = new Parameter( name,
										  units,
										  value);
		}			
		return parameter;
	}	
	
	/**
	 * extract the species from the list and return them in the return-array
	 * 
	 * @param list
	 * @return
	 */
	private static Species[] extractSpecies(List list) throws Exception
	{
		if (list == null)
			return null;
		
		String name;
		String compartment;		
		String units; 
		
		double initialAmount;
		
		int charge;
		
		boolean boundaryCondition;
		
		Species[] species = new Species[list.size()];
		
		for (int i = 0; i < species.length; i++) 
		{
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.NAME) != null)
				name = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.NAME);
			else
			{
				if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.ID) != null)
					name = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.ID);
				else
					name = "";
			}
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.COMPARTMENT) != null)
				compartment = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.COMPARTMENT);
			else
				compartment = "";			
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.UNITS) != null)
				units = ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.UNITS);
			else
				units = "moles";		
			
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.INITIAL_AMOUNT) != null)
				initialAmount = Double.parseDouble( ((Element)list.get(i)).getAttributeValue(SBML_Identifiers.INITIAL_AMOUNT));
			else
				initialAmount = Double.NaN;						
			
			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.CHARGE) != null)
				charge = Integer.parseInt(((Element)list.get(i)).getAttributeValue(SBML_Identifiers.CHARGE));
			else
				charge = Integer.MIN_VALUE;	

			if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.BOUNDARY_CONDITION) != null)
				if (((Element)list.get(i)).getAttributeValue(SBML_Identifiers.BOUNDARY_CONDITION).equals("true"))
					boundaryCondition = true;
				else
					boundaryCondition = false;				
			else
				boundaryCondition = false;			
			
			species[i] = new Species( name,
									  compartment,
									  units,
									  initialAmount,
									  boundaryCondition,
									  charge );
		}			
		return species;
	}		
}
